{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# {class}`tkinter.Canvas` 示例\n",
    "\n",
    "**画布**（canvas）小部件管理着 2D 图形对象（lines, circles, images, 其他更多小部件）组成的集合。Canvas 小部件是经典的 Tk 小部件，不是 Ttk 小部件。\n",
    "\n",
    "创建方法是：\n",
    "\n",
    "```python\n",
    "from tkinter import Canvas\n",
    "canvas = Canvas(parent)\n",
    "```\n",
    "\n",
    "注意：在 Canvas 中的坐标系是以左上角作为原点 $(0,0)$，水平向右为 $x$ 轴正方向，垂直向下为 $y$ 轴正方向。\n",
    "\n",
    "## 创建线段\n",
    "\n",
    "创建线段是以 $(x_0, y_0, x_1, y_1)$ 的形式传入 `creata_line` 函数的。其中 $(x_0, y_0)$，$(x_1, y_1)$ 分别为起点和终点。比如：\n",
    "\n",
    "```python\n",
    "item_id = canvas.create_line(10, 10, 200, 50)\n",
    "```\n",
    "函数 `creata_line` 的返回值 `item_id` 是一个整数，被用来作为该对象的引用的唯一标识。\n",
    "\n",
    "下面看一个例子：\n",
    "\n",
    "```python\n",
    "class Segment(Canvas):\n",
    "    def __init__(self, master=None, **kw):\n",
    "        super().__init__(master=master, **kw)\n",
    "        self.lastx, self.lasty = 0, 0\n",
    "        self.bind(\"<Button-1>\", self.xy) # 绑定鼠标左键\n",
    "        self.bind(\"<B1-Motion>\", self.add_line) # 拖动鼠标左键\n",
    "        \n",
    "    def xy(self, event):\n",
    "        '''更新坐标'''\n",
    "        self.lastx, self.lasty = event.x, event.y\n",
    "        \n",
    "    def add_line(self, event):\n",
    "        '''画一条线段'''\n",
    "        self.create_line(self.lastx, self.lasty, event.x, event.y, fill='red', width=3)\n",
    "        self.xy(event)\n",
    "    \n",
    "    def layout(self):\n",
    "        self.grid(column=0, row=0, sticky='nwes')\n",
    "\n",
    "root = Tk()\n",
    "root.columnconfigure(0, weight=1)\n",
    "root.rowconfigure(0, weight=1)\n",
    "segment = Segment(root)\n",
    "segment.layout()\n",
    "root.mainloop()\n",
    "```\n",
    "该例子实现了在画布上拖动鼠标左键来画线段的目标。其中的参数 `width` 表示线段的宽度，`fill` 表示使用的画笔的颜色。\n",
    "\n",
    "可以做如下修改：\n",
    "\n",
    "```python\n",
    "from tkinter import IntVar\n",
    "\n",
    "class App(Tk):\n",
    "    def __init__(self):\n",
    "        super().__init__()\n",
    "        self.columnconfigure(0, weight=1)\n",
    "        self.rowconfigure(0, weight=1)\n",
    "        self.segment = Segment(self)\n",
    "        self.id_var = IntVar()\n",
    "        \n",
    "    def modify(self):\n",
    "        item_id = self.id_var.get()\n",
    "        self.segment.itemconfigure(item_id, fill='blue', width=10)\n",
    "        \n",
    "    def layout(self):\n",
    "        entry = ttk.Entry(textvariable=self.id_var)\n",
    "        button = ttk.Button(text='modify', command=self.modify)\n",
    "        self.segment.layout()\n",
    "        entry.grid(column=0, row=1)\n",
    "        button.grid(column=1, row=1)\n",
    "        \n",
    "app = App()\n",
    "app.layout()\n",
    "app.mainloop()\n",
    "```\n",
    "\n",
    "效果图见图1：\n",
    "\n",
    "![图1 Canvas 的一个例子](https://upload-images.jianshu.io/upload_images/1114626-301a9217b641804c.png?imageMogr2/auto-orient/strip%7CimageView2/2/w/1240)\n",
    "\n",
    "只要写入文本框数字，然后点击按钮 `modify`，便可 使用 `Canvas.itemconfigure` 函数修改 `item_id` 对应的点的配置。\n",
    "\n",
    "## 绑定 item_id\n",
    "\n",
    "除了可以使用 `bind` 函数绑定事件之外，还可以使用 `tag_bind` 函数绑定  item_id 来触发事件。\n",
    "\n",
    "下面的代码可以使用鼠标点击颜色选择块来切换画笔的颜色：\n",
    "\n",
    "```python\n",
    "class Segment(Canvas):\n",
    "    def __init__(self, master=None, **kw):\n",
    "        super().__init__(master=master, **kw)\n",
    "        self.lastx, self.lasty = 0, 0\n",
    "        self.color = \"black\"\n",
    "        self.bind(\"<Button-1>\", self.xy)\n",
    "        self.bind(\"<B1-Motion>\", self.add_line)\n",
    "        \n",
    "    def set_color(self, new_color):\n",
    "        self.color = new_color\n",
    "    \n",
    "    def xy(self, event):\n",
    "        '''更新坐标'''\n",
    "        self.lastx, self.lasty = event.x, event.y\n",
    "        \n",
    "    def add_line(self, event):\n",
    "        self.create_line((self.lastx, self.lasty, event.x, event.y), fill=self.color)\n",
    "        self.xy(event)\n",
    "\n",
    "    def change(self):\n",
    "        # 创建 3 个颜色选择块\n",
    "        red_id = self.create_rectangle((10, 10, 30, 30), fill=\"red\")\n",
    "        blue_id = self.create_rectangle((10, 35, 30, 55), fill=\"blue\")\n",
    "        black_id = self.create_rectangle((10, 60, 30, 80), fill=\"black\")\n",
    "        # 绑定事件\n",
    "        self.tag_bind(red_id, \"<Button-1>\", lambda x: self.set_color(\"red\"))\n",
    "        self.tag_bind(blue_id, \"<Button-1>\", lambda x: self.set_color(\"blue\"))\n",
    "        self.tag_bind(black_id, \"<Button-1>\", lambda x: self.set_color(\"black\"))\n",
    "    \n",
    "root = Tk()\n",
    "seg = Segment(root)\n",
    "seg.grid()\n",
    "seg.change()\n",
    "root.mainloop()\n",
    "```\n",
    "\n",
    "效果图：\n",
    "\n",
    "![图2 可改变颜色的画布](https://upload-images.jianshu.io/upload_images/1114626-e6cbc3f6f5c47ae4.png?imageMogr2/auto-orient/strip%7CimageView2/2/w/1240)\n",
    "\n",
    "## 标记：Tags\n",
    "\n",
    "使用配置选项 `\"tags\"` 可以为任何一个 `item_id` 代表的对象做标记，这里的标记可以是一个或者多个。\n",
    "\n",
    "为什么要做标记呢？这是因为，做了标记的 item_id 将会更方便管理。比如，您可以指定所有被标记为 `\"line\"` 的 item_id 统一修改其画笔颜色。\n",
    "\n",
    "除了使用配置选项 `\"tags\"` 创建标记之外，您可以在 item_id 创建之后使用 `add_tag` 函数创建或者添加新的标记。移除标记可以使用方法 `dtag`。您也可以使用 `gettags(item_id)` 方式获取 item_id 的全部标记列表。可以使用 `find_withtag` 函数获取指定的标记的全部 item_id 列表。\n",
    "\n",
    "直接看一个例子：\n",
    "\n",
    "```python\n",
    "class Segment(Canvas):\n",
    "    def __init__(self, master=None, **kw):\n",
    "        super().__init__(master=master, **kw)\n",
    "        self.lastx, self.lasty = 0, 0\n",
    "        self.color = \"黑色\"\n",
    "        self.bind(\"<Button-1>\", self.xy)\n",
    "        self.bind(\"<B1-Motion>\", self.add_line)\n",
    "        # 设定调色板的线宽\n",
    "        self.itemconfigure('调色板', width=5) \n",
    "        # 释放鼠标触发事件\n",
    "        self.bind('<B1-ButtonRelease>', self.done_stroke)\n",
    "    \n",
    "    @property\n",
    "    def color_map(self):\n",
    "        return {\n",
    "            '红色': 'red',\n",
    "            '蓝色': 'blue',\n",
    "            '黑色': 'black'\n",
    "        }\n",
    "        \n",
    "    def set_color(self, new_color):\n",
    "        self.color = new_color\n",
    "        self.dtag('all', '被选中的调色板')\n",
    "        self.itemconfigure('调色板', outline='white')\n",
    "        self.addtag('被选中的调色板', 'withtag', f\"{self.color}调色板\")\n",
    "        self.itemconfigure('被选中的调色板', outline='#999999')\n",
    "    \n",
    "    def xy(self, event):\n",
    "        '''更新坐标'''\n",
    "        self.lastx, self.lasty = event.x, event.y\n",
    "        \n",
    "    def add_line(self, event):\n",
    "        loc = (self.lastx, self.lasty, event.x, event.y)\n",
    "        color = self.color_map[self.color]\n",
    "        self.create_line(loc, fill=color, width=5, tags='当前的线段')\n",
    "        self.xy(event)\n",
    "        \n",
    "    def palette(self, loc, color):\n",
    "        kw = {\n",
    "            'fill': color,\n",
    "            'tags': ('调色板', f'{color}调色板')\n",
    "        }\n",
    "        return self.create_rectangle(loc, **kw)\n",
    "\n",
    "    def change(self):\n",
    "        # 创建 3 个颜色选择块\n",
    "        red_id = self.palette((10, 10, 30, 30), \"red\")\n",
    "        blue_id = self.palette((10, 35, 30, 55), \"blue\")\n",
    "        black_id = self.palette((10, 60, 30, 80), \"black\")\n",
    "        # 添加标记\n",
    "        self.addtag('被选中的调色板', 'withtag', black_id)\n",
    "        # 绑定事件\n",
    "        self.tag_bind(red_id, \"<Button-1>\", lambda x: self.set_color(\"红色\"))\n",
    "        self.tag_bind(blue_id, \"<Button-1>\", lambda x: self.set_color(\"蓝色\"))\n",
    "        self.tag_bind(black_id, \"<Button-1>\", lambda x: self.set_color(\"黑色\"))\n",
    "        \n",
    "    def done_stroke(self, event):\n",
    "        self.itemconfigure('当前的线段', width=1)\n",
    "    \n",
    "root = Tk()\n",
    "seg = Segment(root)\n",
    "seg.grid()\n",
    "seg.change()\n",
    "root.mainloop()\n",
    "```\n",
    "\n",
    "显示的效果图：\n",
    "\n",
    "![图3 可改变线条的画布](https://upload-images.jianshu.io/upload_images/1114626-e65c77c5f5ed8952.png?imageMogr2/auto-orient/strip%7CimageView2/2/w/1240)\n",
    "\n",
    "该例子展示了一种使用 tags 修改 item 的手段，达到的效果是：在鼠标释放之前线条比较粗，释放之后线条会变细。\n",
    "\n",
    "## 修改 item\n",
    "\n",
    "我们可以使用 `delete` 方法删除 item，使用 `coords` 方法改变 item 的尺寸和位置（允许变换坐标系）。可以使用 `move` 方法移动 item。还有 `\"raise\"` 和 `\"lower\"` 方法可以改变不同画布的排列布局。比如：\n",
    "\n",
    "### 改变  item 属性\n",
    "```python\n",
    "from tkinter import *\n",
    "\n",
    "root=Tk()\n",
    "\n",
    "cv=Canvas(root,bg='white')\n",
    "rt1=cv.create_rectangle(10,10,110,110,tags=('r1','r2','r3'))\n",
    "rt2=cv.create_rectangle(20,20,80,80,tags=('s1','s2','s3'))\n",
    "rt3=cv.create_rectangle(30,30,70,70,tags=('y1','y2','y3'))\n",
    "\n",
    "cv.tag_lower(rt3)\n",
    "cv.tag_raise(rt1)\n",
    "cv.itemconfig(cv.find_above(rt2),outline='red')\n",
    "cv.itemconfig(cv.find_below(rt2),outline='green')\n",
    "cv.pack()\n",
    "\n",
    "root.mainloop()\n",
    "```\n",
    "\n",
    "### 删除  item\n",
    "```\n",
    "root=Tk()\n",
    "\n",
    "cv=Canvas(root,bg='white')\n",
    "rt1=cv.create_rectangle(10,10,110,110,tags=('r1','r2','r3'))\n",
    "rt2=cv.create_rectangle(20,20,80,80,tags=('s1','s2','s3'))\n",
    "rt3=cv.create_rectangle(30,30,70,70,tags=('s1','y2','y3'))\n",
    "\n",
    "cv.delete(rt1)\n",
    "cv.delete('s1')\n",
    "cv.pack()\n",
    "\n",
    "root.mainloop()\n",
    "```\n",
    "\n",
    "### 缩放 item\n",
    "\n",
    "```python\n",
    "root=Tk()\n",
    "cv=Canvas(root,bg='white')\n",
    "rt1=cv.create_rectangle(10,10,110,110,tags=('r1','r2','r3'))\n",
    "cv.scale(rt1,0,0,1,2)\n",
    "cv.pack()\n",
    "\n",
    "root.mainloop()\n",
    "```\n",
    "\n",
    "## 滚动鼠标\n",
    "\n",
    "在许多应用程序中，您希望画布大于屏幕上显示的内容。 您可以通过 \"xview\" 和 \"yview\" 方法以通常的方式将水平和垂直滚动条附加到画布上。\n",
    "\n",
    "至于画布的大小，您既可以指定希望在屏幕上显示的大小，也可以指定需要滚动才能看到的画布的完整大小。 画布小部件的 \"width\" 和 \"height\" 配置选项将从几何管理器请求给定的空间量。\"scrollregion\" 配置选项（例如 \"0 0 1000 1000\"）告诉 Tk 画布表面有多大。\n",
    "\n",
    "针对鼠标滚动，\"canvasx\" 和 \"canvasy\" 方法会将屏幕上的位置（正在报告的绑定）转换为画布上的实际点。如果将它们直接添加到事件绑定中（而不是从事件绑定中调用），请注意引用和替换，以确保在事件触发时完成转换。\n",
    "\n",
    "下面的例子很好的说明了这种机制：\n",
    "\n",
    "```python\n",
    "class SegmentScroll(Segment):\n",
    "    def __init__(self, master=None, **kw):\n",
    "        super().__init__(master=master, **kw)\n",
    "        self.master = master\n",
    "        \n",
    "    def scroll(self):\n",
    "        self._h = ttk.Scrollbar(orient='horizontal')\n",
    "        self._v = ttk.Scrollbar(orient='vertical')\n",
    "        # 告诉 Tk 画布表面有多大\n",
    "        self['scrollregion'] = (0, 0, 1000, 1000)\n",
    "        self.itemconfig('scrollregion', )\n",
    "        self.configure(yscrollcommand=self._v.set, xscrollcommand=self._h.set)\n",
    "        self._h['command'] = self.xview\n",
    "        self._v['command'] = self.yview\n",
    "        \n",
    "    def custom(self):\n",
    "        ttk.Sizegrip(root).grid(column=1, row=1, sticky=(S,E))\n",
    "        self.grid(column=0, row=0, sticky=(N,W,E,S))\n",
    "        self._h.grid(column=0, row=1, sticky=(W,E))\n",
    "        self._v.grid(column=1, row=0, sticky=(N,S))\n",
    "        self.master.grid_columnconfigure(0, weight=1)\n",
    "        self.master.grid_rowconfigure(0, weight=1)\n",
    "        \n",
    "    def xy(self, event):\n",
    "        self.lastx, self.lasty = self.canvasx(event.x), self.canvasy(event.y)\n",
    "        \n",
    "    def add_line(self, event):\n",
    "        x, y = self.canvasx(event.x), self.canvasy(event.y)\n",
    "        color = self.color_map[self.color]\n",
    "        self.create_line((self.lastx, self.lasty, x, y), fill=color, width=5, tags='当前的线段')\n",
    "        self.xy(event)\n",
    "        \n",
    "root = Tk()\n",
    "seg = SegmentScroll(root)\n",
    "seg.change()\n",
    "seg.scroll()\n",
    "seg.custom()\n",
    "root.mainloop()\n",
    "```\n",
    "\n",
    "显示效果：\n",
    "\n",
    "![图4 带滚动条和缩放的画布](https://upload-images.jianshu.io/upload_images/1114626-b376821e69d188af.png?imageMogr2/auto-orient/strip%7CimageView2/2/w/1240)\n",
    "\n",
    "## 更改线段的样式\n",
    "\n",
    "我们也可以设定参数 `arrow` 和 `arrowshape` 来改变线段的样式：\n",
    "\n",
    "```python\n",
    "root = Tk()\n",
    "\n",
    "cv = Meta(root, bg='white')\n",
    "d = [(0, 'none'), (1, 'first'), (2, 'last'), (3, 'both')]\n",
    "for i in d:\n",
    "    cv.create_line((10, 10+i[0]*20, 110, 110+i[0]*20),\n",
    "                   arrow=i[1], arrowshape='40 30 10')\n",
    "\n",
    "cv.grid()\n",
    "root.mainloop()\n",
    "```\n",
    "\n",
    "显示效果图：\n",
    "\n",
    "![图5 画出不同样式的线段](https://upload-images.jianshu.io/upload_images/1114626-f6dd20c14d5cc102.png?imageMogr2/auto-orient/strip%7CimageView2/2/w/1240)\n",
    "\n",
    "还有：\n",
    "\n",
    "```python\n",
    "root = Tk()\n",
    "\n",
    "cv = Canvas(root, bg='white')\n",
    "d = [(0, 'none', 'bevel'), (1, 'first', 'miter'),\n",
    "     (2, 'last', 'round'), (3, 'both', 'round')]\n",
    "for i in d:\n",
    "    cv.create_line((10, 10+i[0]*20, 110, 110+i[0]*20),\n",
    "                   arrow=i[1], arrowshape='8 10 3', joinstyle=i[2])\n",
    "\n",
    "cv.grid()\n",
    "root.mainloop()\n",
    "```\n",
    "\n",
    "效果图：\n",
    "\n",
    "![图6 直线的 joinstyle 参数设定](https://upload-images.jianshu.io/upload_images/1114626-e029829c6324fdf5.png?imageMogr2/auto-orient/strip%7CimageView2/2/w/1240)\n",
    "\n",
    "\n",
    "## 画布的其他设定\n",
    "\n",
    "Canvas 除了支持 \"line\", \"rectangle\"，还支持 \"oval\", \"arc\", \"polygon\",  \"bitmap\" (位图，可用于分割物体), \"image\",\"text\"，甚至还支持 \"window\" 的嵌入。\n",
    "\n",
    "### 绘制位图\n",
    "\n",
    "```python\n",
    "root = Tk()\n",
    "self = Canvas(root, background='white')\n",
    "bitmap = ('error', 'info', 'question', 'hourglass',\n",
    "          \"warning\", \"gray12\",\n",
    "         \"gray25\", \"gray50\", \"gray75\", \"questhead\")\n",
    "\n",
    "for k, name in enumerate(bitmap):\n",
    "    location = [20*(k+1)]*2\n",
    "    self.create_bitmap(location, bitmap=name)\n",
    "    \n",
    "self.grid()\n",
    "root.mainloop()\n",
    "```\n",
    "\n",
    "![图7 位图的绘制](https://upload-images.jianshu.io/upload_images/1114626-aba646939c9f5439.png?imageMogr2/auto-orient/strip%7CimageView2/2/w/1240)\n",
    "\n",
    "### 绘制多边形\n",
    "\n",
    "```python\n",
    "root = Tk()\n",
    "self = Canvas(root)\n",
    "# 点的坐标\n",
    "points = (10, 10), (10, 200), (90, 200), (200, 160)\n",
    "self.create_polygon(points, fill='red')\n",
    "    \n",
    "self.grid()\n",
    "root.mainloop()\n",
    "```\n",
    "\n",
    "![图8 绘制多边形](https://upload-images.jianshu.io/upload_images/1114626-c7c284d44652f2da.png?imageMogr2/auto-orient/strip%7CimageView2/2/w/1240)\n",
    "\n",
    "### 绘制文本\n",
    "\n",
    "```python\n",
    "root = Tk()\n",
    "self = Canvas(root)\n",
    "location = 50, 50\n",
    "text = self.create_text(location, text='一个文本：永不言败！', anchor='sw', fill='blue', font='italic 15')\n",
    "# 选中文本\n",
    "self.select_from(text, 5)\n",
    "self.select_to(text, 8)\n",
    "self.grid()\n",
    "root.mainloop()\n",
    "```\n",
    "\n",
    "![图9 绘制文本](https://upload-images.jianshu.io/upload_images/1114626-1fdaf1f808ac9d1c.png?imageMogr2/auto-orient/strip%7CimageView2/2/w/1240)\n",
    "\n",
    "### 创建组件\n",
    "\n",
    "```python\n",
    "from tkinter import Canvas, ttk, Tk\n",
    "\n",
    "root = Tk()\n",
    "self = Canvas(root)\n",
    "def print_text():\n",
    "    print(\"你好\")\n",
    "    \n",
    "bt = ttk.Button(self, text='点我', command=print_text)\n",
    "self.create_window((10, 10), window=bt, anchor='w')\n",
    "self.create_line(30, 30, 50, 90)\n",
    "self.grid()\n",
    "root.mainloop()\n",
    "```\n",
    "\n",
    "![图10 创建组件](https://upload-images.jianshu.io/upload_images/1114626-87363502921b065f.png?imageMogr2/auto-orient/strip%7CimageView2/2/w/1240)\n",
    "\n"
   ]
  }
 ],
 "metadata": {
  "language_info": {
   "name": "python"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 2
}
